# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

from enum import Enum


class ReportFormat(str, Enum):
    JSON = "JSON"
    CSV = "CSV"

    def __str__(self):
        return str(self.value)



import numpy as np

from .. import job as job_schema
from .. import projected


def test_projected_gaussian():
    xs = np.arange(-10, 10, 0.1)
    assert projected.gaussian(xs)[0] == 1.9287498479639178e-22


def test_get_potential_from_weights():
    assert projected.get_potential_from_weights(
        list(range(1, 122, 1)), [1.0, 2.2, 3.5]
    ) == [85.91219226992594, 87.25507420353789, 88.46888252900956]


def test_get_corrected_projection_weights():
    bar = job_schema.Barrier()
    res = projected.get_corrected_projection_weights(bar.get_ideal_potential, time=2.0)
    assert len(res) == 121
    assert res[30] == 1.5881403774524704e-272


def test_get_actual_potential():
    bar = job_schema.Barrier()
    res = projected.get_actual_potential(
        bar.get_ideal_potential, time=2.0, positions=[1.0, 2.2, 3.5]
    )
    assert res == [6.631842134176605, 9.772009614670178, 4.038341506097124]



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Literal

from pydantic import BaseModel, ConfigDict


class QuestionnaireName(str, Enum):
    REGISTRATION = "Registration"
    ACCT_UPGRADE = "Account Upgrade"

    def __str__(self):
        return str(self.value)


class UseReason(str, Enum):
    """
    # TECH_USE = "Technology Development"
    # SCI_USE = "Scientific Research"
    # EDU_USE = "Education"
    # CUR_USE = "Curious"
    # OTH_USE = "Other" """

    SCI_USE = "SCI_USE"
    TECH_USE = "TECH_USE"
    EDU_USE = "EDU_USE"
    CUR_USE = "CUR_USE"
    OTH_USE = "OTH_USE"

    def __str__(self):
        return str(self.value)


class SchoolLevel(str, Enum):
    """
    # PRIMARY = "Primary (K-8)"
    # SECONDARY = "Secondary (9-12)"
    # COLLEGE = "College/University"
    # POSTGRAD = "Post-Graduate"
    # NOTGIVEN = "Not Given"
    PRIMARY = "PRIMARY" """

    SECONDARY = "SECONDARY"
    COLLEGE = "COLLEGE"
    POSTGRAD = "POSTGRAD"
    NOTGIVEN = "NOTGIVEN"

    def __str__(self):
        return str(self.value)


class RegistrationQandA(BaseModel):
    uses: list[UseReason]
    school_level: SchoolLevel | None = SchoolLevel.NOTGIVEN
    model_config = ConfigDict(validate_assignment=True, use_enum_values=True)


class AcctUpgradeQandA(BaseModel):
    uses: list[UseReason]
    more: str | None = None
    model_config = ConfigDict(validate_assignment=True, use_enum_values=True)


class Questionnaire(BaseModel):
    name: QuestionnaireName
    q_and_a: RegistrationQandA | AcctUpgradeQandA
    model_config = ConfigDict(validate_assignment=True, use_enum_values=True)


class RegistrationQuestionnaire(Questionnaire):
    name: Literal[QuestionnaireName.REGISTRATION] = QuestionnaireName.REGISTRATION
    q_and_a: RegistrationQandA


class AcctUpgradeQuestionnaire(Questionnaire):
    name: Literal[QuestionnaireName.ACCT_UPGRADE] = QuestionnaireName.ACCT_UPGRADE
    q_and_a: AcctUpgradeQandA



# type: ignore
import warnings
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Annotated
from uuid import UUID

import numpy as np
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    StringConstraints,
    computed_field,
    conlist,
    field_validator,
    model_validator,
)
from scipy.interpolate import interp1d
from typing_extensions import TypedDict

from .projected import (
    BarrierHeight,
    BarrierHeights,
    BarrierPosition,
    BarrierPositions,
    BarrierTime,
    BarrierTimes,
    BarrierWidth,
    BarrierWidths,
    ProjectedEnergies,
    ProjectedPositions,
    get_actual_potential,
)
from .projected import Settings as ProjectionSettings
from .qpu import QPUBase, QPUName
from .timing import (
    EndTimeMs,
    TimeMs,
    TimeOfFlightMs,
    decimal_times_are_unique,
    decimal_to_float,
    decimals_to_floats,
    float_to_decimal,
    resolve_times,
)
from .timing import Settings as TimeSettings

pset = ProjectionSettings()
tset = TimeSettings()

JobName = Annotated[str, StringConstraints(min_length=1, max_length=50)]

JobNote = Annotated[str, StringConstraints(max_length=500)]


class JobOrigin(str, Enum):
    WEB = "WEB"
    OQTAPI = "OQTAPI"


class JobType(str, Enum):
    BEC = "BEC"
    BARRIER = "BARRIER"
    BRAGG = "BRAGG"
    ATOMTRONIC_1D = "ATOMTRONIC_1D"
    PAINT_1D = "PAINT_1D"

    def __str__(self):
        return str(self.value)


class ImageType(str, Enum):
    IN_TRAP = "IN_TRAP"
    TIME_OF_FLIGHT = "TIME_OF_FLIGHT"
    MOT = "MOT"

    def __str__(self):
        return str(self.value)


class ImageAppId(str, Enum):
    tof_imaging = "TOF imaging"
    mot_imaging = "MOT imaging"
    it_imaging = "IT imaging"


class FitMethod(str, Enum):
    gaussian_fit_2d = "Gaussian Fit 2D"
    bimodal_fit = "Bimodal Fit 2D"
    bimodal_fit_zero_temperature = "Bimodal Fit 2D (zero temperature)"
    barrier_1d = "Barrier 1D"


class OutputJobType(str, Enum):
    IN_TRAP = "IN_TRAP"
    NON_IN_TRAP = "NON_IN_TRAP"

    def __str__(self):
        return str(self.value)


class JobStatus(str, Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    COMPLETE = "COMPLETE"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"
    INCOMPLETE = "INCOMPLETE"

    def __str__(self):
        return str(self.value)


class RfInterpolationType(str, Enum):
    LINEAR = "LINEAR"
    OFF = "OFF"
    PREVIOUS = "PREVIOUS"  # assumes value of previous data point

    def __str__(self):
        return str(self.value)


class InterpolationType(str, Enum):
    LINEAR = "LINEAR"
    SMOOTH = "SMOOTH"
    OFF = "OFF"
    # native scipy options
    ZERO = "ZERO"  # spline interpolation at zeroth order
    SLINEAR = "SLINEAR"  # spline interpolation at first order
    QUADRATIC = "QUADRATIC"  # spline interpolation at second order
    CUBIC = "CUBIC"  # spline interpolation at third order
    # LINEAR = "LINEAR"         # self explanatory
    NEAREST = "NEAREST"  # assumes value of nearest data point
    PREVIOUS = "PREVIOUS"  # assumes value of previous data point
    NEXT = "NEXT"  # assumes value of next data point

    def __str__(self):
        return str(self.value)


class LaserType(str, Enum):
    TERMINATOR = "TERMINATOR"
    BRAGG = "BRAGG"

    def __str__(self):
        return str(self.value)


class ShapeType(str, Enum):
    GAUSSIAN = "GAUSSIAN"
    LORENTZIAN = "LORENTZIAN"
    SQUARE = "SQUARE"
    NATIVE = "NATIVE"

    def __str__(self):
        return str(self.value)


def interpolation_to_kind(interpolation: InterpolationType) -> str:
    """Method to convert our InterpolationType to something scipy can understand

    Args:
        interpolation (bert_schemas.job.InterpolationType): Primitive job interpolation type

    Returns:
        str: A "kind" string to be used by scipy's interp1d
    """
    interpolation_map = {"OFF": "zero", "SMOOTH": "cubic"}

    return interpolation_map.get(interpolation, interpolation.lower())


def interpolate_1d(
    xs: list[float],
    ys: list[float],
    x: float,
    interpolation: InterpolationType = "LINEAR",
) -> float:
    """Method to interpolate a 1D list of pairs [xs, ys] at the evaluation point x

    Args:
        xs (list[float]): List of x values
        ys (list[float]): List of y values
        x (float): Desired x-coordinate to evaluate the resulting interpolation function
        interpolation (job_schema.InterpolationType, optional): Interpolation style

    Returns:
        float: Interpolation function value at the specified x-coordinate
    """
    f = interp1d(
        xs,
        ys,
        kind=interpolation_to_kind(interpolation),
        bounds_error=False,
        fill_value=(0.0, 0.0),
    )
    return f(x)[()]  # extract value


def interpolate_1d_list(
    xs: list[float],
    ys: list[float],
    x_values: list[float],
    interpolation: InterpolationType = "LINEAR",
) -> list[float]:
    """Method to interpolate a 1d list of pairs [xs, ys] at the evaluation points given by x_values

    Args:
        xs (list[float]): List of x values
        ys (list[float]): List of y values
        x_values (list[float]): Desired x-coordinates to evaluate the resulting interpolation function
        interpolation (job_schema.InterpolationType, optional): Interpolation style

    Returns:
        list[float]: Floating point values corresponding to evaluation of the interpolation function
            value at the specified x_values
    """
    f = interp1d(
        xs,
        ys,
        kind=interpolation_to_kind(interpolation),
        bounds_error=False,
        fill_value=(0.0, 0.0),
    )
    return list(f(x_values))


class AbsorptionImages(BaseModel):
    atoms: np.ndarray = Field(
        default_factory=lambda: np.zeros(shape=(200, 200))
    )  # Absolute Image
    back: np.ndarray = Field(
        default_factory=lambda: np.zeros(shape=(200, 200))
    )  # Background Image
    dark: np.ndarray = Field(default_factory=lambda: np.zeros(shape=(200, 200)))

    class Config:
        arbitrary_types_allowed = True


class CroppedCloud(BaseModel):
    od_image: np.ndarray = Field(default_factory=lambda: np.zeros(shape=(100, 100)))
    found: bool = False
    x_center: int = 0
    y_center: int = 0

    class Config:
        arbitrary_types_allowed = True


class Image(BaseModel):
    pixels: list[float]
    rows: int
    columns: int
    pixcal: float | None = 1.0
    model_config = ConfigDict(validate_assignment=True, extra="forbid")


class RawImages(BaseModel):
    image_type: ImageAppId
    absolute: Image
    background: Image | None


class RawImagesResponse(RawImages):
    job_external_id: UUID
    run: int


class Point(TypedDict):
    x: float
    y: float


class LineChart(BaseModel):
    points: list[dict[str, float]]
    model_config = ConfigDict(validate_assignment=True, extra="forbid")


class RfEvaporation(BaseModel):
    # times_ms upper range can be no larger than end_time_ms of job (80 ms is upper default)
    times_ms: Annotated[
        list[
            Annotated[
                int,
                Field(ge=-tset.max_evap_duration, le=tset.max_exp_time),
            ]
        ],
        Field(min_length=1, max_length=20),
    ] = list(range(-1600, 400, 400))
    frequencies_mhz: Annotated[
        list[Annotated[float, Field(ge=0.0, le=25.0)]],
        Field(min_length=1, max_length=20),
    ]
    powers_mw: Annotated[
        list[Annotated[float, Field(ge=0.0, le=1000.0)]],
        Field(min_length=1, max_length=20),
    ]
    interpolation: RfInterpolationType
    model_config = ConfigDict(validate_assignment=True, extra="forbid")

    @field_validator("times_ms", mode="before")
    @classmethod
    def times_are_unique(cls, v):
        if not all(np.diff(v)):
            raise ValueError("Time values must be unique.")
        return v

    @field_validator("times_ms", mode="before")
    @classmethod
    def two_values_less_equal_zero(cls, v):
        if not len([elem for elem in v if elem <= 0]) >= 2:
            raise ValueError("At least two time values must be <= 0.")
        return v

    @model_validator(mode="after")
    def cross_validate(self) -> "RfEvaporation":
        if not len(self.times_ms) == len(self.frequencies_mhz) == len(self.powers_mw):
            raise ValueError("RfEvaporation data lists must have the same length.")

        if self.times_ms != sorted(self.times_ms):
            warnings.warn(
                "Evaporation times_ms list must be naturally ordered, re-ordering.",
                stacklevel=2,
            )
            self.times_ms, self.frequencies_mhz, self.powers_mw = zip(
                *sorted(
                    zip(
                        self.times_ms,
                        self.frequencies_mhz,
                        self.powers_mw,
                    )
                )
            )
        return self


class Landscape(BaseModel):
    # time_ms upper range can be no larger than end_time_ms of job (80 ms is upper default)
    time_ms: TimeMs = Decimal("0.0")
    potentials_khz: ProjectedEnergies = [0.0, 0.0]
    positions_um: ProjectedPositions = [-1.0, 1.0]
    spatial_interpolation: InterpolationType = InterpolationType.LINEAR

    @property
    def interpolation_kind(self) -> str:
        if self.spatial_interpolation == InterpolationType.OFF:
            kind = "zero"
        else:
            kind = InterpolationType[self.spatial_interpolation].name.lower()
        return kind

    @model_validator(mode="after")
    def cross_validate(self):
        if not len(self.positions_um) == len(self.potentials_khz):
            raise ValueError("Landscape data lists must have the same length.")

        if self.positions_um != sorted(self.positions_um):
            warnings.warn(
                "Landscape positions_um list must be naturally ordered, re-ordering.",
                stacklevel=2,
            )
            self.positions_um, self.potentials_khz = zip(
                *sorted(zip(self.positions_um, self.potentials_khz))
            )
        return self

    def get_ideal_potential(
        self, time: float = 0.0, positions: list = pset.projected_spots
    ) -> list[float]:
        """Method to get the ideal potential energy at the specified positions

        Args:
            positions (list, optional): List of positions in microns

        Returns:
            list[float]: Potential energies, in kHz, at the specified positions
        """
        potentials = interpolate_1d_list(
            self.positions_um,
            self.potentials_khz,
            positions,
            self.spatial_interpolation,
        )
        return potentials

    def get_potential(self, positions: list[float]) -> list[float]:
        """Method to calculate the optical potential associated with a Landscape object,
        taking into account the actual implementation of the Oqtant projection system,
        at the given time

        Args:
            positions (list[float]): Positions, in microns, where the potential should be evaluated

        Returns:
            list[float]: Potential energies, in kHz, at the specified positions
        """
        return get_actual_potential(self.get_ideal_potential, positions=positions)

    def get_position_spectrum(self, positions_um: np.ndarray) -> np.ndarray:
        """Get positional weights over the given positions"""
        spectrum = interpolate_1d_list(
            self.positions_um,
            self.potentials_khz,
            positions_um,
            self.interpolation_kind,
        )
        return np.asarray(spectrum)

    def __lt__(self, other):
        return self.time_ms < other.time_ms


class OpticalLandscape(BaseModel):
    interpolation: InterpolationType = InterpolationType.LINEAR
    landscapes: Annotated[list[Landscape], Field(min_length=2, max_length=5)]
    model_config = ConfigDict(validate_assignment=True)

    @property
    def interpolation_kind(self) -> str:
        if self.interpolation == "OFF":
            kind = "zero"
        else:
            kind = InterpolationType[self.interpolation].name.lower()
        return kind

    @model_validator(mode="after")
    def cross_validate(self):
        # ensure the individual Landscape objects are far enough apart in time and naturally (time) ordered
        if sorted(self.landscapes) != self.landscapes:
            self.landscapes = sorted(
                self.landscapes, key=lambda landscape: landscape.time_ms
            )
        ts = [landscape.time_ms for landscape in self.landscapes]
        if not decimal_times_are_unique(ts):
            raise ValueError(
                "Constituent Landscape object time_ms values must differ by >= {pset.update_period} ms."
            )
        return self

    def get_ideal_potential(
        self, time: float, positions: list[float] = pset.projected_spots
    ) -> list[float]:
        """Method to calculate ideal object potential energy at the specified time and positions

        Args:
            time (float): Time, in ms, at which the potential energy is calculated
            positions (list[float], optional): Positions at which the potential energy is calculated

        Returns:
            list[float]: Potential energies, in kHz, at specified time and positions
        """
        potential = [0.0] * len(positions)
        snaps = self.landscapes if hasattr(self, "landscapes") else self.snapshots

        if len(snaps) < 2:
            return potential
        snap_times = [decimal_to_float(snap.time_ms) for snap in snaps]
        if time >= min(snap_times) and time <= max(snap_times):
            pre = next(
                snap
                for snap in reversed(snaps)
                if decimal_to_float(snap.time_ms) <= time
            )
            nex = next(snap for snap in snaps if decimal_to_float(snap.time_ms) >= time)
            potential = [
                interpolate_1d(
                    list(map(float, [pre.time_ms, nex.time_ms])),
                    [p1, p2],
                    time,
                    self.interpolation,
                )
                for p1, p2 in zip(
                    pre.get_ideal_potential(positions=positions),
                    nex.get_ideal_potential(positions=positions),
                )
            ]
        return potential

    def get_potential(
        self, time: float, positions: list = pset.projected_spots
    ) -> list[float]:
        """Method to calculate the optical potential associated with a Landscape object,
        taking into account the actual implementation of the Oqtant projection system,
        at the given time

        Args:
            time (float): Time, in ms, at which to sample the potential energy
            positions (list[float], optional): Positions, in microns, where the potential should be evaluated

        Returns:
            list[float]: Potential energies, in kHz, at the requested positions and time
        """
        return get_actual_potential(
            self.get_ideal_potential, time=time, positions=positions
        )

    def get_position_spectra(
        self, times_ms: list, positions_um: np.ndarray
    ) -> np.ndarray:
        """Get the position spectrum over the specified positions at the given times"""
        potentials_khz = np.zeros(shape=(len(times_ms), len(positions_um)))
        # construct "current" profile as weighted sum of constituent profiles
        profiles = self.landscapes
        for indx, time_ms in enumerate(times_ms):
            is_active = time_ms >= decimal_to_float(
                profiles[0].time_ms
            ) and time_ms < decimal_to_float(profiles[-1].time_ms)
            if not is_active:
                potentials_khz[indx] = np.zeros_like(potentials_khz[0])
            else:
                prev_profile = next(
                    profile
                    for profile in reversed(profiles)
                    if decimal_to_float(profile.time_ms) <= time_ms
                )
                next_profile = next(
                    profile
                    for profile in profiles
                    if decimal_to_float(profile.time_ms) > time_ms
                )
                prev_potential = prev_profile.get_position_spectrum(positions_um)
                next_potential = next_profile.get_position_spectrum(positions_um)
                t_prev = decimal_to_float(prev_profile.time_ms)
                t_next = decimal_to_float(next_profile.time_ms)
                prev_weight = (t_next - time_ms) / (t_next - t_prev)
                next_weight = (time_ms - t_prev) / (t_next - t_prev)
                # snapshot landscapes/profiles connected w/ linear interpolation in time
                potentials_khz[indx] = (
                    prev_weight * prev_potential + next_weight * next_potential
                )
        return potentials_khz


class TofFit(BaseModel):
    gaussian_od: float
    gaussian_sigma_x: float
    gaussian_sigma_y: float
    tf_od: float
    tf_x: float
    tf_y: float
    x_0: float
    y_0: float
    offset: float
    model_config = ConfigDict(validate_assignment=True, extra="forbid")


class Barrier(BaseModel):
    # times_ms upper range can be no larger than end_time_ms of job (80 ms is upper default)
    times_ms: BarrierTimes = list(map(Decimal, list(np.arange(1, 12, 1.0))))
    positions_um: BarrierPositions = list(np.arange(1, 12, 1.0))
    heights_khz: BarrierHeights = [10.0] * 11
    widths_um: BarrierWidths = [1.0] * 11
    interpolation: InterpolationType = InterpolationType.LINEAR
    shape: ShapeType = ShapeType.GAUSSIAN
    model_config = ConfigDict(validate_assignment=True)

    @property
    def interpolation_kind(self) -> str:
        if self.interpolation == "OFF":
            kind = "zero"
        else:
            kind = InterpolationType[self.interpolation].name.lower()
        return kind

    @model_validator(mode="before")
    def native_widths(data):
        """data can be a Barrier object or a dict"""
        obj = False
        if hasattr(data, "shape"):
            obj = True
            data = data.model_dump()

        if data.get("shape") == ShapeType.NATIVE:
            if "widths_um" in data:
                if not all(x == 0.5 for x in data["widths_um"]):
                    raise ValueError(
                        "NATIVE shaped Barrier object widths must all be equal to 0.5"
                    )
            elif "heights_khz" in data:
                data["widths_um"] = [0.5] * len(data["heights_khz"])
            else:
                data["widths_um"] = [0.5] * 11
        if obj:
            data = Barrier(**data)
        return data

    @field_validator("times_ms", mode="before")
    @classmethod
    def times_are_unique(cls, v):
        if not decimal_times_are_unique(v):
            raise ValueError("Time values must be unique.")
        return v

    @model_validator(mode="after")
    def cross_validate(self):
        if not (
            len(self.times_ms)
            == len(self.positions_um)
            == len(self.heights_khz)
            == len(self.widths_um)
        ):
            raise ValueError("Barrier data lists must have the same length.")

        if self.times_ms != sorted(self.times_ms):
            warnings.warn(
                "Barrier times_ms list must be naturally ordered, re-ordering.",
                stacklevel=2,
            )
            (
                self.times_ms,
                self.positions_um,
                self.heights_khz,
                self.widths_um,
            ) = zip(
                *sorted(
                    zip(
                        self.times_ms,
                        self.positions_um,
                        self.heights_khz,
                        self.widths_um,
                    )
                )
            )

        return self

    @property
    def lifetime(self) -> float:
        """Property to get the lifetime value of a Barrier object

        Returns:
            float: The amount of time, in ms, that the barrier will exist
        """
        return self.death - self.birth

    @property
    def birth(self) -> float:
        """Property to get the (manipulation stage) time that the Barrier object will be created

        Returns:
            float: The time, in ms, at which the barrier will start being projected
        """
        return min(decimals_to_floats(self.times_ms))

    @property
    def death(self) -> float:
        """Property to get the (manipulation stage) time that the Barrier object will cease to exist

        Returns:
            float: The time, in ms, at which the barrier will stop being projected
        """
        return max(decimals_to_floats(self.times_ms))

    @property
    def is_precision(self) -> bool:
        """Property to ask if this barrier is a "precision" one, i.e. one with the narrowest possible
        width at all times and the smoothest possible dynamics.

        Returns:
            bool: True if the Barrier object is classified as precision, False otherwise
        """
        # return bool(
        #     all([width <= pset.min_barrier_width for width in self.widths_um])
        #     and self.shape == ShapeType.GAUSSIAN
        # )
        return self.shape == ShapeType.NATIVE

    def evolve(
        self,
        duration: Annotated[BarrierTime, Field(ge=pset.update_period)],
        position: BarrierPosition = None,
        height: BarrierHeight = None,
        width: BarrierWidth = None,
    ) -> None:
        """Method to evolve the position, height, and/or width of a Barrier object over a duration

        Args:
            duration (TimeMs): The time, in ms, over which evolution should take place
            position (ProjectedPosition | None, optional): The position, in microns, to evolve to
            height (ProjectedEnergy | None, optional): The height, in kHz, to evolve to
            width (BarrierWidth | None, optional): The width, in microns, to evolve to
        """
        if position is None:
            position = self.positions_um[-1]
        if height is None:
            height = self.heights_khz[-1]
        if width is None:
            width = self.widths_um[-1]
        self.positions_um.append(position)
        self.heights_khz.append(height)
        self.widths_um.append(width)
        self.times_ms.append(self.times_ms[-1] + float_to_decimal(duration))

    def is_active(self, time: float) -> bool:
        """Method to determine if a Barrier object is active (exists) at the specified time

        Args:
            time (float): The time, in ms, at which the query is evaluated

        Returns:
            bool: Flag indicating if the barrier exists or not at the specified time
        """
        return time >= decimal_to_float(self.times_ms[0]) and time <= decimal_to_float(
            self.times_ms[-1]
        )

    def get_positions(self, times: list[float]) -> list[float]:
        """Method to calculate the Barrier object position at the specified (manipulation stage) times

        Args:
            times (list[float]): The times, in ms, at which positions are calculated

        Returns:
            list[float]: The positions, in microns, at the specified times
        """
        return interpolate_1d_list(
            decimals_to_floats(self.times_ms),
            self.positions_um,
            resolve_times(times),
            self.interpolation_kind,
        )

    def get_position(self, time: float) -> float:
        """Method to calculate the Barrier object position at the specified (manipulation stage) time

        Args:
            time (float): The time, in ms, at which the position is calculated

        Returns:
            float: The position, in microns, at the specified time
        """
        return self.get_positions(times=[time])[0]

    def get_heights(self, times: list[float]) -> list[float]:
        """Method to calculate the Barrier object heights at the specified list of times

        Args:
            times (list[float]): The times, in ms, at which the heights are calculated

        Returns:
            list[float]: The barrier heights at the specified times
        """
        return interpolate_1d_list(
            decimals_to_floats(self.times_ms),
            self.heights_khz,
            resolve_times(times),
            self.interpolation_kind,
        )

    def get_height(self, time: float) -> float:
        """Method to get the Barrier object height at the specified time

        Args:
            time (float): The time, in ms, at which the height is calculated

        Returns:
            float: The barrier height at the specified time
        """
        return self.get_heights(times=[time])[0]

    def get_widths(self, times: list[float]) -> list[float]:
        """Method to calculate the Barrier object widths at the specified times

        Args:
            times (list[float]): The times, in ms, at which the heights are calculated

        Returns:
            list[float]: The barrier widths at the specified times
        """
        return interpolate_1d_list(
            decimals_to_floats(self.times_ms),
            self.widths_um,
            resolve_times(times),
            self.interpolation_kind,
        )

    def get_width(self, time: float) -> float:
        """Method to calculate the Barrier object width at the specified time

        Args:
            times (float): The time, in ms, at which the height is calculated

        Returns:
            float: The barrier width at the specified time
        """
        return self.get_widths(times=[time])[0]

    def get_params(self, time_ms: float) -> dict:
        kind = self.get_interpolation_kind()
        params = {}
        params["position_um"] = interpolate_1d(
            decimals_to_floats(self.times_ms), self.positions_um, time_ms, kind
        )
        params["width_um"] = interpolate_1d(
            decimals_to_floats(self.times_ms), self.widths_um, time_ms, kind
        )
        params["height_khz"] = interpolate_1d(
            decimals_to_floats(self.times_ms), self.heights_khz, time_ms, kind
        )
        return params

    def get_ideal_potential(
        self, time: float = 0.0, positions: list[float] = pset.projected_spots
    ) -> list[float]:
        """Method to calculate the ideal Barrier object potential energy at the given positions
        and at the specified time without taking into account finite projection system resolution
        to update time of projected light

        Args:
            time (float, optional): The time, in ms, at which the potential is calculated
            positions (list[float], optional): The positions, in microns, at which the potential
                energies are evaluated

        Returns:
            list[float]: The potential energies, in kHz, at the specified positions
        """
        height = self.get_height(time)
        position = self.get_position(time)
        weight = self.get_width(time)
        potential = [0.0] * len(positions)
        if height <= 0 or weight <= 0 or not self.is_active(time):
            return potential
        if self.shape == "SQUARE":  # width = half width
            potential = [
                0 if (p < position - weight or p > position + weight) else height
                for p in positions
            ]
        elif self.shape == "LORENTZIAN":  # width == HWHM (half-width half-max)
            potential = [
                height / (1 + ((p - position) / weight) ** 2) for p in positions
            ]
        elif self.shape in ("GAUSSIAN", "NATIVE"):  # width = sigma (Gaussian width)
            potential = [
                height * np.exp(-((p - position) ** 2) / (2 * weight**2))
                for p in positions
            ]
        return potential

    def get_potential(
        self, time: float, positions: list[float] = pset.projected_spots
    ) -> list[float]:
        """Method to calculate the optical potential associated with a Barrier object, taking into
        account the actual implementation of the Oqtant projection system

        Args:
            time (float): The time, in ms, at which the potential should be evaluated
            positions (list[float], optional): The positions, in microns, at which the potential should be evaluated

        Returns:
            list[float]: The potential energies, in kHz, at the specified positions
        """
        return get_actual_potential(
            self.get_ideal_potential, time=time, positions=positions
        )

    def evaluate_position_spectrum(
        self, h: float, x0: float, w: float, positions_um: list[float]
    ) -> list[float]:
        if h == 0.0 or w == 0.0:
            return np.zeros_like(positions_um)
        if self.shape == "SQUARE":
            # width = half width (to align more closely with other shapes)
            return np.asarray(
                [0 if (x < x0 - w or x > x0 + w) else h for x in positions_um]
            )
        elif self.shape == ShapeType.LORENTZIAN:
            # width == HWHM (half-width half-max)
            return np.asarray([h / (1 + ((x - x0) / w) ** 2) for x in positions_um])
        elif self.shape in (ShapeType.GAUSSIAN, ShapeType.NATIVE):
            # width = sigma (Gaussian width)
            return np.asarray(
                [h * np.exp(-((x - x0) ** 2) / (2 * w**2)) for x in positions_um]
            )

    def get_position_spectrum(
        self, time_ms: float, positions_um: np.ndarray
    ) -> np.ndarray:
        """Get positional weights over the given positions at the specified time"""
        params = self.get_params(time_ms)
        h = params["height_khz"]
        x0 = params["position_um"]
        w = params["width_um"]
        if h > 0 and w > 0 and self.is_active(time_ms):
            return self.evaluate_position_spectrum(h, x0, w, positions_um)
        else:
            return np.zeros(len(positions_um))

    def get_position_spectra(self, times_ms: list, positions_um: list) -> np.ndarray:
        widths = self.get_widths(times_ms)
        heights = self.get_heights(times_ms)
        positions = self.get_positions(times_ms)

        amplitudes = np.zeros(shape=(len(times_ms), len(positions_um)), dtype=float)
        for indx in range(len(times_ms)):
            amplitudes[indx] = self.evaluate_position_spectrum(
                h=heights[indx],
                w=widths[indx],
                x0=positions[indx],
                positions_um=positions_um,
            )
        return amplitudes


class Pulse(BaseModel):
    # times_ms upper range can be no larger than end_time_ms of job (MAX_USER_TIME_MS ms is upper default)
    times_ms: Annotated[list[TimeMs], Field(min_length=2, max_length=2)] = [0, 1]
    intensities_mw_per_cm2: Annotated[
        list[Annotated[float, Field(ge=1.0, le=1.0)]], Field(min_length=2, max_length=2)
    ] = [1, 1]
    detuning_mhz: Annotated[float, Field(ge=0, le=0)] = 0
    interpolation: InterpolationType = InterpolationType.OFF
    model_config = ConfigDict(validate_assignment=True, extra="forbid")

    @model_validator(mode="before")
    @classmethod
    def cross_validate(cls, values):
        if not len(values["times_ms"]) == len(values["intensities_mw_per_cm2"]):
            raise ValueError("Pulse data lists must have the same length.")
        return values

    @field_validator("times_ms", mode="before")
    @classmethod
    def times_are_unique(cls, v):
        if not decimal_times_are_unique(v):
            raise ValueError("Time values must be unique.")
        return v

    @field_validator("detuning_mhz", mode="before")
    @classmethod
    def resonant_detuning_check(cls, v):
        if not v == 0:
            raise ValueError("Only resonant (detuning=0) pulses are supported.")
        return v

    @field_validator("intensities_mw_per_cm2", mode="before")
    @classmethod
    def default_intensity(cls, v):
        for i in v:
            if not i == 1:
                raise ValueError("Intensity control is not supported.")
        return v

    @field_validator("times_ms", mode="before")
    @classmethod
    def pulse_timing(cls, v):
        if not len(v) == 2:
            raise ValueError("Pulses must be defined by at exactly two times.")
        if not decimal_times_are_unique(v):
            raise ValueError("Pulse times must be unique.")
        if not v == sorted(v):
            raise ValueError("Pulse times must be naturally ordered.")
        return v

    @field_validator("interpolation", mode="before")
    @classmethod
    def no_interpolation(cls, v):
        if not v == InterpolationType.OFF:
            raise ValueError("Interpolation not supported for laser pulses.")
        return v

    def __lt__(self, other):
        return min(self.times_ms) < min(other.times_ms)


class Laser(BaseModel):
    type: LaserType = LaserType.TERMINATOR
    position_um: float = pset.terminator_position
    pulses: Annotated[list[Pulse], Field(min_length=1, max_length=1)]
    model_config = ConfigDict(validate_assignment=True)

    @field_validator("pulses", mode="before")
    @classmethod
    def pulse_timing(cls, v):
        if len(v) > 1:
            raise ValueError("Only a single laser pulse is currently supported.")
        for index, _ in enumerate(v):
            if index < len(v) - 1:
                dt_ms = min(v[index + 1].times_ms) - max(v[index].times_ms)
                if not decimal_times_are_unique([dt_ms]):
                    raise ValueError(
                        "Pulse overlap; end of one pulse and start of the next must differ by >= {pset.update_period} ms."
                    )
        return v

    @field_validator("position_um", mode="before")
    @classmethod
    def default_position(cls, v):
        if not v == pset.terminator_position:
            raise ValueError("Setting position for laser pulses is not supported.")
        return v

    def get_intensity_waveform(
        self, tstart_ms: float, tend_ms: float, sample_rate_hz: float
    ) -> np.ndarray:
        intensities = []
        time_ms = tstart_ms
        for pulse in self.pulses:
            # pad our intensities list with zeros until the start of the next pulse:
            dt = 1.0 / sample_rate_hz
            dt_ms = dt * 1000.0
            n_zeros = int(
                np.floor(
                    ((decimal_to_float(pulse.times_ms[0]) - time_ms) / 1000.0)
                    * sample_rate_hz
                )
            )
            if n_zeros > 0:
                intensities.extend(np.zeros(n_zeros))
            # jump to start time of pulse and interpolate over it
            time_ms += n_zeros * 1000.0 * dt
            times_ms = np.arange(time_ms, decimal_to_float(pulse.times_ms[-1]), dt_ms)
            style = (
                "zero"
                if pulse.interpolation == "OFF"
                else InterpolationType[pulse.interpolation].name.lower()
            )
            f = interp1d(
                decimals_to_floats(pulse.times_ms),
                pulse.intensities_mw_per_cm2,
                kind=style,
                bounds_error=False,
                fill_value=(0.0, 0.0),
                assume_sorted=True,
                copy=False,
            )
            intensities.extend(f(times_ms))
            # jump current time to the end of the pulse
            time_ms += len(times_ms) * dt_ms
        # extend intensities list to the desired end time
        n_zeros = int(np.floor(((tend_ms - time_ms) / 1000.0) * sample_rate_hz))
        if n_zeros > 0:
            intensities.extend(np.zeros(n_zeros))
        return np.array(intensities)

    @property
    def detunings(self) -> list[float]:
        return [pulse.detuning_mhz for pulse in self.pulses]

    @property
    def detuning_triggers(self) -> list[float]:
        trigger_times_ms = [
            pulse.times_ms[-1] for pulse in self.pulses
        ]  # get last pulse
        trigger_times_ms.insert(0, 0)
        return trigger_times_ms[: len(self.pulses)]

    def is_on(self, time_ms) -> bool:
        for pulse in self.pulses:
            if (time_ms >= decimal_to_float(pulse.times_ms[0])) & (
                time_ms <= decimal_to_float(pulse.times_ms[-1])
            ):
                return True
        return False


class NonPlotOutput(BaseModel):
    mot_fluorescence_image: Image
    tof_image: Image
    tof_fit_image: Image
    tof_fit: TofFit
    tof_x_slice: LineChart
    tof_y_slice: LineChart
    total_mot_atom_number: int
    tof_atom_number: int
    thermal_atom_number: int
    condensed_atom_number: int
    temperature_nk: int
    model_config = ConfigDict(
        from_attributes=True, validate_assignment=True, extra="forbid"
    )


class PlotOutput(BaseModel):
    it_plot: Image
    model_config = ConfigDict(
        from_attributes=True, validate_assignment=True, extra="forbid"
    )


class Output(BaseModel):
    input_id: int | None = None
    values: PlotOutput | NonPlotOutput
    model_config = ConfigDict(validate_assignment=True)


class JobOutput(Output): ...


class BecOutput(Output):
    values: NonPlotOutput


class BarrierOutput(Output):
    values: NonPlotOutput | PlotOutput


class InputValues(BaseModel):
    end_time_ms: EndTimeMs
    image_type: ImageType
    time_of_flight_ms: TimeOfFlightMs
    rf_evaporation: RfEvaporation
    optical_barriers: (
        Annotated[list[Barrier], Field(min_length=1, max_length=10)] | None
    ) = None
    optical_landscape: OpticalLandscape | None = None
    lasers: Annotated[list[Laser], Field(min_length=1, max_length=1)] | None = None

    @model_validator(mode="after")
    def cross_validate(self):
        if list(
            filter(
                lambda time_ms: time_ms > decimal_to_float(self.end_time_ms),
                self.rf_evaporation.times_ms,
            )
        ):
            raise ValueError(
                "rf_evaporation.times_ms max values cannot exceed end_time_ms"
            )
        if self.optical_barriers:
            for index, optical_barrier in enumerate(self.optical_barriers):
                if list(
                    filter(
                        lambda time_ms: decimal_to_float(time_ms)
                        > decimal_to_float(self.end_time_ms),
                        optical_barrier.times_ms,
                    )
                ):
                    raise ValueError(
                        f"optical_barriers[{index}].times_ms max values cannot exceed end_time_ms"
                    )
        if self.optical_landscape:
            for index, landscape in enumerate(self.optical_landscape.landscapes):
                if landscape.time_ms > self.end_time_ms:
                    raise ValueError(
                        f"optical_landscape.landscapes[{index}].time_ms max value cannot exceed end_time_ms"
                    )
        if self.lasers:
            for laser_index, laser in enumerate(self.lasers):
                for pulse_index, pulse in enumerate(laser.pulses):
                    if list(
                        filter(
                            lambda time_ms: time_ms > self.end_time_ms,
                            pulse.times_ms,
                        )
                    ):
                        raise ValueError(
                            f"lasers[{laser_index}].pulses[{pulse_index}].times_ms max values cannot exceed end_time_ms"
                        )
        return self

    @field_validator("lasers")
    @classmethod
    def check_lasers(cls, v):
        if v:
            if len(v) > 1:
                raise ValueError("Multiple lasers not supported.")
            for laser in v:
                if not laser.type == LaserType.TERMINATOR:
                    raise ValueError(
                        "Must be type TERMINATOR. Other laser types not supported."
                    )
        return v


class Input(BaseModel):
    job_id: int | None = None
    run: int | None = None
    values: InputValues
    output: Output | None = None
    notes: JobNote | None = None
    model_config = ConfigDict(validate_assignment=True)


class RfEvaporationWithoutValidation(BaseModel):
    times_ms: list[float]
    frequencies_mhz: list[float]
    powers_mw: list[float]
    interpolation: RfInterpolationType


class OpticalBarriersWithoutValidation(BaseModel):
    times_ms: list[float]
    positions_um: list[float]
    heights_khz: list[float]
    widths_um: list[float]
    interpolation: InterpolationType
    shape: ShapeType


class LandscapeWithoutValidation(BaseModel):
    time_ms: float
    potentials_khz: list[float]
    positions_um: list[float]
    spatial_interpolation: InterpolationType


class OpticalLandscapeWithoutValidation(BaseModel):
    interpolation: InterpolationType
    landscapes: list[LandscapeWithoutValidation]


class PulseWithoutValidation(BaseModel):
    times_ms: list[float]
    intensities_mw_per_cm2: list[float]
    detuning_mhz: float
    interpolation: InterpolationType


class LaserWithoutValidation(BaseModel):
    type: LaserType
    position_um: float
    pulses: list[PulseWithoutValidation]


class InputValuesWithoutValidation(BaseModel):
    end_time_ms: float
    image_type: ImageType
    time_of_flight_ms: float
    rf_evaporation: RfEvaporationWithoutValidation
    optical_barriers: list[OpticalBarriersWithoutValidation] | None = None
    optical_landscape: OpticalLandscapeWithoutValidation | None = None
    lasers: list[LaserWithoutValidation] | None = None


class InputWithoutOutput(Input):
    values: InputValuesWithoutValidation
    output: object | None = Field(exclude=True)


class JobBase(BaseModel):
    name: JobName
    origin: JobOrigin | None = None
    status: JobStatus = JobStatus.PENDING
    display: bool = True
    time_start: datetime | None = None
    time_complete: datetime | None = None
    qpu_name: QPUName = QPUName.UNDEFINED
    input_count: int | None = None
    inputs: conlist(
        Input,
        min_length=1,
        max_length=30,
    )

    @computed_field
    @property
    def job_type(self) -> JobType:
        input_values = self.inputs[0].values
        if input_values.optical_landscape:
            return JobType.PAINT_1D
        elif (
            input_values.optical_barriers
            or input_values.image_type == ImageType.IN_TRAP
        ):
            return JobType.BARRIER
        else:
            return JobType.BEC

    @model_validator(mode="after")
    def input_run_populate(self):
        for i, _ in enumerate(self.inputs):
            if not self.inputs[i].run:
                self.inputs[i].run = i + 1
        return self

    model_config = ConfigDict(use_enum_values=True, validate_assignment=True)


# needed for post fixtures
class JobPost(JobBase):
    inputs: conlist(
        Input,
        min_length=1,
        max_length=30,
    )


class JobCreate(JobBase):
    @model_validator(mode="after")
    def compute_input_count(self):
        if self.input_count != len(self.inputs):
            self.input_count = len(self.inputs)
        return self


class ResponseInput(BaseModel):
    job_id: int | None = None
    run: int | None = None
    values: InputValuesWithoutValidation
    output: JobOutput | None = None
    notes: JobNote | None = None
    model_config = ConfigDict(from_attributes=True)


class JobResponse(JobBase):
    external_id: UUID
    qpu: QPUBase | None = None
    time_submit: datetime
    paid: bool = False
    inputs: list[ResponseInput]
    failed_inputs: list[int] = []


class JobInputsResponse(JobResponse):
    qpu_name: QPUName = QPUName.UNDEFINED
    raw_images: bool = False
    inputs: list[InputWithoutOutput]


class PaginatedJobsResponse(JobInputsResponse):
    external_id: UUID
    time_submit: datetime
    time_start: datetime | None = None
    time_complete: datetime | None = None
    model_config = ConfigDict(from_attributes=True, use_enum_values=True)


class Job(JobBase):
    job_id: UUID


class ExternalId(BaseModel):
    id: UUID


class UpdateJobDisplay(BaseModel):
    job_external_id: UUID
    display: bool = True
    model_config = ConfigDict(validate_assignment=True, extra="forbid")


class JobCreateResponse(BaseModel):
    job_id: UUID
    queue_position: int
    est_time: int | str


class JobExternalIdsList(BaseModel):
    external_ids: list[UUID]



from .. import timing
from decimal import Decimal


def test_resolve_time():
    assert timing.resolve_time(3.11) == 3.1


def test_resolve_times():
    assert timing.resolve_times([3.11, 3.21]) == [3.1, 3.2]


def test_settings():
    settings = timing.Settings()
    assert settings.max_evap_duration == 2000


def test_decimal_times_are_unique():
    times = [1.1, 1.2]
    assert timing.decimal_times_are_unique(times)


def test_decimal_times_are_not_unique():
    times = [1.1, 1.15]
    assert not timing.decimal_times_are_unique(times)


def test_decimal_to_float():
    assert timing.decimal_to_float(Decimal("1.1")) == 1.1


def test_decimals_to_floats():
    assert timing.decimals_to_floats([Decimal("1.25"), Decimal("1.3")]) == [1.2, 1.3]


def test_float_to_decimal():
    assert timing.float_to_decimal(76.54) == Decimal("76.5")


def test_floats_to_decimals():
    assert timing.floats_to_decimals([76.54, 23.56]) == [
        Decimal("76.5"),
        Decimal("23.5"),
    ]



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module that captures the features, and limitations, of optical objects
implemented by the Oqtant hardware projection system.
"""


from collections.abc import Callable

import numpy as np
from typing import Annotated
from pydantic import Field
from .timing import resolve_time, TimeMs
from decimal import Decimal
from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class Settings:
    resolution: float = 2.2  # 1/e^2 diameter of projection system, microns
    position_step: float = 1.0  # grid step between projected spots, microns
    min_position: float = -60.0  # minimum position of projected light, microns
    max_position: float = 60  # maximum position of projected light, microns
    num_projected_spots: int = 121
    update_period: Decimal = Decimal(
        "0.1"
    )  # milliseconds between updates of projected light
    max_energy: float = 100.0  # maximum projected energy shift at any position, kHz
    min_barrier_width: float = 0.5
    max_barrier_width: float = 50
    max_num_barrier_times: int = 20
    max_num_barriers: int = 20
    terminator_position: float = (
        30.0  # position of the center of the gaussian terminator beam in um
    )
    terminator_width: float = 20.7  # 1/e^2 width of the terminator beam in um

    @property
    def projected_spots(self) -> list[float]:
        return np.arange(
            self.min_position,
            self.max_position + self.position_step,
            self.position_step,
        )


settings = Settings()


ProjectedPosition = Annotated[
    float, Field(ge=settings.min_position, le=settings.max_position)
]

ProjectedPositions = Annotated[
    list[ProjectedPosition],
    Field(min_length=2, max_length=settings.num_projected_spots),
]

ProjectedEnergy = Annotated[float, Field(ge=0.0, le=settings.max_energy)]

ProjectedEnergies = Annotated[
    list[ProjectedEnergy], Field(min_length=2, max_length=settings.num_projected_spots)
]

BarrierTime = TimeMs

BarrierTimes = Annotated[
    list[BarrierTime], Field(min_length=2, max_length=settings.max_num_barrier_times)
]

BarrierPosition = ProjectedPosition

BarrierPositions = Annotated[
    ProjectedPositions, Field(max_length=settings.max_num_barriers)
]

BarrierHeight = ProjectedEnergy

BarrierHeights = Annotated[
    ProjectedEnergies, Field(max_length=settings.max_num_barrier_times)
]

BarrierWidth = Annotated[
    float, Field(ge=settings.min_barrier_width, le=settings.max_num_barrier_times)
]

BarrierWidths = Annotated[
    list[BarrierWidth], Field(min_length=2, max_length=settings.max_num_barrier_times)
]

BarrierTimes = Annotated[
    list[TimeMs], Field(min_length=2, max_length=settings.max_num_barrier_times)
]


def get_potential_from_weights(weights: list, positions: list[float]) -> list[float]:
    """Method to calculate projected potential based on given weights (intensities)
    applied to each projected spot

    Args:
        weights: height of each Gaussian spot

    Returns:
        list[float]: Calculated total optical potential at the given positions
    """
    if len(weights) != settings.num_projected_spots:
        raise ValueError(
            f"number of weights values must equal {settings.num_projected_spots}"
        )

    potential = np.zeros_like(positions, dtype=np.float64)
    for indx, spot in enumerate(settings.projected_spots):
        potential += gaussian(
            xs=positions,
            amp=weights[indx],
            center=spot,
            sigma=settings.resolution / 4.0,
            offset=0.0,
        )
    return list(potential)


def get_corrected_projection_weights(
    get_ideal_potential: Callable[[float], list], time: float = 0
) -> np.ndarray:
    """Method to calculate weights for each horizontal "spot" projected onto the atom ensemble to
    attempt to achieve the passed optical object's "ideal" potential energy profile.
    Implements first-order corrections for anamolous contributions from nearby spots,
    inter-integer barrier centers, etc

    Args:
        get_ideal_potential (Callable[[float], list]): Method for the optical object or any class
            that supports optical objects that calculates the specified "ideal" or "requested"
            potential energy profile
        time (float, optional): Time at which to correct

    Returns:
        np.ndarray[float]: Calculated (optical intensity) contribution for each projected spot
            (diffraction frequency) used by the projection systems
    """
    bin_size = 10  # index range of local scaling window
    positions_fine = np.arange(
        settings.min_position - settings.position_step / 2,
        settings.max_position + settings.position_step / 2,
        settings.position_step / bin_size,
    )

    # determine the ideal potential energy at each projected spot
    potential_ideal = np.asarray(
        get_ideal_potential(time=time, positions=positions_fine)
    )
    potential_ideal_course = np.asarray(
        get_ideal_potential(time=time, positions=settings.projected_spots)
    )

    # calculate the optical field that would result from raw object data
    potential_raw = get_potential_from_weights(
        weights=potential_ideal_course, positions=positions_fine
    )
    # bin them up and find the local maximum
    potential_ideal_binned = np.asarray(potential_ideal).reshape(
        int(len(list(potential_raw)) / bin_size), bin_size
    )
    potential_raw_binned = np.asarray(potential_raw).reshape(
        int(len(list(potential_raw)) / bin_size), bin_size
    )
    # scale local optical potential to get close to the ideal value
    maxes_ideal = np.asarray([np.max(x) for x in potential_ideal_binned])
    maxes_raw = np.asarray([np.max(x) for x in potential_raw_binned])
    scalings = np.divide(
        maxes_ideal,
        maxes_raw,
        where=maxes_raw != 0.0,
        out=np.zeros_like(maxes_ideal, dtype=np.float64),
    )
    return list(np.multiply(scalings, potential_ideal_course))


def get_actual_potential(
    get_ideal_potential: Callable[[float], list],
    time: float = 0.0,
    positions: list = settings.projected_spots,
) -> list[float]:
    """Method to calculate the "actual" potential energy vs position for optical
    objects/fields as realized by the Oqtant projection system. Includes effects,
    and first-order corrections for, finite time updates and finite optical
    resolution/optical objects being projected as sums of gaussians and energetic
    clipping of optical potentials

    Args:
        get_ideal_potential (Callable[[float], list]): Object method for request/ideal potential
        time (float, optional): Time to evaluate ideal potential
        positions (list[float], optional): Positions to evaluate the actual potential at

    Returns:
        list[float]: Expected actual potential energy at the request positions
    """
    weights = get_corrected_projection_weights(
        get_ideal_potential, time=resolve_time(time)
    )

    corrected_potential = get_potential_from_weights(
        weights=weights,
        positions=positions,
    )
    return list(np.clip(np.asarray(corrected_potential), 0.0, settings.max_energy))


def gaussian(
    xs: np.ndarray,
    amp: float = 1.0,
    center: float = 0.0,
    sigma: float = 1.0,
    offset: float = 0.0,
) -> np.ndarray:
    """Method that evaluates a standard gaussian form over the given input points

    Args:
        xs (numpy.ndarray): Positions where the gaussian should be evaluated
        amp (float, optional): Gaussian amplitude
        center (float, optional): Gaussian center
        sigma (float, optional): Gaussian width
        offset (float, optional): Gaussian dc offset

    Returns:
        np.ndarray: Gaussian function evaluated over the input points
    """
    return amp * np.exp(-((xs - center) ** 2) / (2 * sigma**2)) + offset



import numpy as np
import pytest
from bert_schemas import job as job_schema
from bert_schemas.job import interpolate_1d, interpolate_1d_list
from pydantic import ValidationError


@pytest.fixture
def barrier_data():
    yield {
        "times_ms": [3, 4],
        "positions_um": [2, 3],
        "heights_khz": [2, 3],
        "widths_um": [2, 3],
        "interpolation": "LINEAR",
    }


@pytest.fixture
def laser_data():
    yield {
        "type": "TERMINATOR",
        "position_um": 30,
        "pulses": [
            {
                "times_ms": [1, 3],
                "intensities_mw_per_cm2": [1, 1],
                "detuning_mhz": 0,
                "interpolation": "OFF",
            }
        ],
    }


@pytest.fixture
def pulse_data():
    yield {
        "times_ms": [1, 2],
        "intensities_mw_per_cm2": [1, 1],
        "detuning_mhz": 0,
        "interpolation": "OFF",
    }


@pytest.fixture
def inputs():
    data = {
        "end_time_ms": 5,
        "image_type": "IN_TRAP",
        "time_of_flight_ms": 2,
        "rf_evaporation": {
            "times_ms": [-3, -4],
            "frequencies_mhz": [2, 1],
            "powers_mw": [2, 3],
            "interpolation": "LINEAR",
        },
        "optical_barriers": [
            {
                "times_ms": [1, 3, 4, 5],
                "positions_um": [0, 2, 3, 6],
                "heights_khz": [2, 3, 4, 5],
                "widths_um": [1, 2, 3, 4],
                "interpolation": "LINEAR",
            }
        ],
        "optical_landscape": {
            "interpolation": "LINEAR",
            "landscapes": [
                {
                    "time_ms": 3,
                    "potentials_khz": [2, 3],
                    "positions_um": [2, 3],
                    "spatial_interpolation": "LINEAR",
                },
                {
                    "time_ms": 4,
                    "potentials_khz": [2, 3],
                    "positions_um": [2, 3],
                    "spatial_interpolation": "LINEAR",
                },
            ],
        },
        "lasers": [
            {
                "type": "TERMINATOR",
                "position_um": 30,
                "pulses": [
                    {
                        "times_ms": [1, 3],
                        "intensities_mw_per_cm2": [1, 1],
                        "detuning_mhz": 0,
                        "interpolation": "OFF",
                    }
                ],
            }
        ],
    }
    input_values = job_schema.InputValues(**data)
    yield job_schema.Input(job_id=1, values=input_values)


def test_interpolate_1d():
    assert interpolate_1d([1, 2, 3], [10, 20, 30], 2) == 20


def test_interpolate_1d_list():
    assert interpolate_1d_list([1, 2, 3], [10, 20, 30], [2, 3]) == [20, 30]


def test_rf_evaporation_cross_validate():
    data = {
        "times_ms": [-3, -4],
        "frequencies_mhz": [2, 1],
        "powers_mw": [2, 3],
        "interpolation": "LINEAR",
    }
    assert job_schema.RfEvaporation.model_validate(data)
    data["times_ms"] = [-3, 4, 5]
    with pytest.raises(ValidationError):
        job_schema.RfEvaporation.model_validate(data)


def test_landscape_interpolation_kind():
    data = {
        "time_ms": 3,
        "potentials_khz": [2, 3],
        "positions_um": [2, 3],
        "spatial_interpolation": "LINEAR",
    }
    landscape = job_schema.Landscape(**data)
    assert landscape.interpolation_kind == "linear"

    data["spatial_interpolation"] = "OFF"
    landscape = job_schema.Landscape(**data)
    assert landscape.interpolation_kind == "zero"


def test_landscape_cross_validate():
    data = {
        "time_ms": 3,
        "potentials_khz": [2, 3],
        "positions_um": [2, 3],
        "spatial_interpolation": "LINEAR",
    }
    assert job_schema.Landscape.model_validate(data)

    data["positions_um"] = [2, 3, 4]
    with pytest.raises(ValidationError):
        data["positions_um"] = [2, 3, 4]
        job_schema.Landscape.model_validate(data)


def test_landscape_get_position_spectrum():
    data = {
        "time_ms": 3,
        "potentials_khz": [2, 3],
        "positions_um": [2, 3],
        "spatial_interpolation": "LINEAR",
    }
    landscape = job_schema.Landscape(**data)
    spectrum = landscape.get_position_spectrum(np.array([6, 3]))
    assert list(spectrum) == [0, 3]


def test_landscape_get_potential():
    data = {
        "time_ms": 3,
        "potentials_khz": [2, 3],
        "positions_um": [2, 3],
        "spatial_interpolation": "LINEAR",
    }
    landscape = job_schema.Landscape(**data)
    res = landscape.get_potential([1.0, 2.0, 3.0])
    assert res == [0.2914965035541314, 1.9804369402077469, 2.772117736811176]


def test_optical_landscape_get_potential():
    data = {
        "landscapes": [
            {
                "time_ms": 3,
                "potentials_khz": [2, 3],
                "positions_um": [2, 3],
                "spatial_interpolation": "LINEAR",
            },
            {
                "time_ms": 4,
                "potentials_khz": [2, 3],
                "positions_um": [2, 3],
                "spatial_interpolation": "LINEAR",
            },
        ]
    }

    landscape = job_schema.OpticalLandscape(**data)

    time = 2.0
    positions = [1.0, 2.2, 3.5]
    potentials = landscape.get_potential(time, positions)

    assert len(potentials) == len(positions)


def test_optical_landscape_cross_validate():
    data = {
        "landscapes": [
            {
                "time_ms": 3,
                "potentials_khz": [2, 3],
                "positions_um": [2, 3],
                "spatial_interpolation": "LINEAR",
            },
            {
                "time_ms": 3,
                "potentials_khz": [2, 3],
                "positions_um": [2, 3],
                "spatial_interpolation": "LINEAR",
            },
        ]
    }
    with pytest.raises(ValidationError):
        job_schema.OpticalLandscape.model_validate(data)

    data["landscapes"][1]["time_ms"] = 4
    assert job_schema.OpticalLandscape.model_validate(data)


def test_optical_landscape_get_position_spectra():
    data = {
        "landscapes": [
            {
                "time_ms": 3,
                "potentials_khz": [2, 3],
                "positions_um": [2, 3],
                "spatial_interpolation": "LINEAR",
            },
            {
                "time_ms": 4,
                "potentials_khz": [2, 3],
                "positions_um": [2, 3],
                "spatial_interpolation": "LINEAR",
            },
        ]
    }

    landscape = job_schema.OpticalLandscape(**data)
    spectrum = landscape.get_position_spectra(
        times_ms=np.array([6, 3]), positions_um=np.array([6, 3])
    )
    assert all(list(spectrum)[0]) == all([0.0, 0.0])


def test_native_barrier(barrier_data):
    barrier_data["shape"] = "NATIVE"
    del barrier_data["widths_um"]
    job_schema.Barrier.model_validate(barrier_data)
    barrier = job_schema.Barrier(**barrier_data)
    assert barrier.widths_um == [0.5, 0.5]


def test_native_barrier_equal_widths_error(barrier_data):
    barrier_data["shape"] = "NATIVE"
    with pytest.raises(ValidationError):
        job_schema.Barrier.model_validate(barrier_data)


def test_barrier_cross_validate(barrier_data):
    assert job_schema.Barrier.model_validate(barrier_data)

    barrier_data["heights_khz"] = [2, 3, 4]
    with pytest.raises(ValidationError):
        job_schema.Barrier.model_validate(barrier_data)


def test_barrier_interpolation_kind(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    assert barrier.interpolation_kind == "linear"

    barrier_data["interpolation"] = "OFF"
    barrier = job_schema.Barrier(**barrier_data)
    assert barrier.interpolation_kind == "zero"


def test_barrier_is_active(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    assert barrier.is_active(3)


def test_barrier_get_positions():
    data = {
        "times_ms": [1, 3, 4, 5],
        "positions_um": [0, 2, 3, 6],
        "heights_khz": [2, 3, 4, 5],
        "widths_um": [1, 2, 3, 4],
        "interpolation": "LINEAR",
    }
    barrier = job_schema.Barrier(**data)
    assert list(barrier.get_positions([3, 4])) == [2, 3]


def test_barrier_get_position(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    assert barrier.get_position(3) == 2


def test_barrier_get_heights(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    assert list(barrier.get_heights([3, 4])) == [2, 3]


def test_barrier_get_height(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    assert barrier.get_height(3) == 2


def test_barrier_get_widths(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    assert barrier.get_widths([3, 4]) == [2.0, 3.0]


def test_barrier_get_width(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    assert barrier.get_width(3) == 2.0


def test_barrier_get_ideal_potential(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    assert barrier.get_ideal_potential(3)[0] == 4.1930635165770337e-209


def test_barrier_get_ideal_potential_native(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    barrier.widths_um = [0.5] * len(barrier.widths_um)
    assert barrier.get_ideal_potential(3)[50] == 1.6757885067638739e-125


def test_barrier_get_potential(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    assert barrier.get_potential(3)[0] == 7.070114771075559e-207


def test_barrier_evaluate_position_spectrum(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)

    barrier.shape = job_schema.ShapeType.SQUARE
    spectrum = barrier.evaluate_position_spectrum(1.0, 2.0, 3.0, [6.0, 3.0])
    assert spectrum[1] == 1

    barrier.shape = job_schema.ShapeType.GAUSSIAN
    spectrum = barrier.evaluate_position_spectrum(1.0, 2.0, 3.0, [6.0, 3.0])
    assert spectrum[0] == 0.41111229050718745

    barrier.widths_um = [0.5] * len(barrier.widths_um)
    barrier.shape = job_schema.ShapeType.NATIVE
    spectrum = barrier.evaluate_position_spectrum(1.0, 2.0, 3.0, [6.0, 3.0])
    assert spectrum[0] == 0.41111229050718745


def test_barrier_evolve(barrier_data):
    barrier = job_schema.Barrier(**barrier_data)
    barrier.evolve(duration=0.1, position=10.0, height=20.0, width=2.0)


# need Noah input
# def test_barrier_get_position_spectra(barrier_data):
#     barrier = job_schema.Barrier(**barrier_data)
#     spectra = barrier.get_position_spectra([2, 3], [6, 3])
#     assert spectra == np.ndarray([[0.0, 0.0], [0.27067057, 1.76499381]])


def test_pulse_cross_validate(pulse_data):
    assert job_schema.Pulse.model_validate(pulse_data)

    pulse_data["times_ms"] = [1, 3, 4, 5]
    with pytest.raises(ValidationError):
        job_schema.Pulse.model_validate(pulse_data)


def test_pulse_resonant_detuning_check(pulse_data):
    pulse_data["detuning_mhz"] = 2
    with pytest.raises(ValidationError):
        job_schema.Pulse.model_validate(pulse_data)


def test_pulse_default_intensity(pulse_data):
    pulse_data["intensities_mw_per_cm2"] = [2, 0]
    with pytest.raises(ValidationError):
        job_schema.Pulse.model_validate(pulse_data)


def test_pulse_timing(pulse_data):
    pulse_data["times_ms"] = [1]
    with pytest.raises(ValidationError):
        job_schema.Pulse.model_validate(pulse_data)

    pulse_data["times_ms"] = [1.1, 1.19]
    with pytest.raises(ValidationError):
        job_schema.Pulse.model_validate(pulse_data)

    pulse_data["times_ms"] = [8, 2]
    with pytest.raises(ValidationError):
        job_schema.Pulse.model_validate(pulse_data)


def test_pulse_no_interpolation(pulse_data):
    pulse_data["interpolation"] = "LINEAR"
    with pytest.raises(ValidationError):
        job_schema.Pulse.model_validate(pulse_data)


def test_laser_pulse_timing():
    pulse_data_1 = {
        "times_ms": [1, 3],
        "intensities_mw_per_cm2": [1, 1],
        "detuning_mhz": 0,
        "interpolation": "OFF",
    }

    pulse_data_2 = {
        "times_ms": [2, 3],
        "intensities_mw_per_cm2": [1, 1],
        "detuning_mhz": 0,
        "interpolation": "OFF",
    }

    laser_data = {
        "type": "TERMINATOR",
        "position_um": 30,
        "pulses": [job_schema.Pulse(**pulse_data_1), job_schema.Pulse(**pulse_data_2)],
    }

    with pytest.raises(ValidationError):
        job_schema.Laser.model_validate(laser_data)


def laser_default_position(laser_data):
    laser_data["position_um"] = 1
    with pytest.raises(ValidationError):
        job_schema.Laser.model_validate(laser_data)


def test_laser_is_on(laser_data):
    laser = job_schema.Laser(**laser_data)

    assert laser.is_on(2) is True
    assert laser.is_on(5) is False


# need Noah input
# def test_laser_get_intensity_waveform():
#     laser_data = {
#         "type": "TERMINATOR",
#         "position_um": 1.0,
#         "pulses": [
#             {
#                 "times_ms": [1, 3, 4],
#                 "intensities_mw_per_cm2": [2, 3, 4],
#                 "detuning_mhz": 2,
#                 "interpolation": "LINEAR",
#             },
#         ],
#     }
#     laser = job_schema.Laser(**laser_data)
#     waveform = laser.get_intensity_waveform(
#         tstart_ms=1.0, tend_ms=3.0, sample_rate_hz=1.0
#     )
#


def test_job_base_input_run_populate(inputs):
    inputs2 = job_schema.Input(job_id=1, values=inputs.values)
    job = job_schema.JobBase(name="test", inputs=[inputs, inputs2])
    assert job.inputs[0].run == 1
    assert job.inputs[1].run == 2


def test_input_values_cross_validate(inputs):
    input_values = inputs.values.model_dump()
    assert job_schema.InputValues.model_validate(input_values)

    with pytest.raises(ValidationError):
        input_values["rf_evaporation"]["times_ms"] = [-3, 6]
        job_schema.InputValues.model_validate(input_values)

    with pytest.raises(ValidationError):
        input_values["optical_barriers"][0]["times_ms"] = [1, 3, 4, 6]
        job_schema.InputValues.model_validate(input_values)

    with pytest.raises(ValidationError):
        input_values["optical_landscape"]["landscapes"][0]["time_ms"] = 6
        job_schema.InputValues.model_validate(input_values)

    with pytest.raises(ValidationError):
        input_values["lasers"][0]["pulses"][0]["times_ms"] = [1, 3, 6]
        job_schema.InputValues.model_validate(input_values)



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from enum import Enum, StrEnum

from pydantic import BaseModel, ConfigDict


class QPUName(str, Enum):
    SMALLBERT = "SMALLBERT"
    BIGBERT = "BIGBERT"
    UNDEFINED = "UNDEFINED"
    SIMULATOR = "SIMULATOR"

    def __str__(self):
        return str(self.value)


class QPUStatus(StrEnum):
    ONLINE = "ONLINE"
    OFFLINE = "OFFLINE"
    TESTING = "TESTING"


class QPUJobType(BaseModel):
    name: str

    model_config = ConfigDict(from_attributes=True)


class QPUAccess(BaseModel):
    day: str
    start_time: time
    end_time: time

    model_config = ConfigDict(from_attributes=True)


class QPUBase(BaseModel):
    name: QPUName
    status: QPUStatus
    model_config = ConfigDict(from_attributes=True)


class QPU(QPUBase):
    job_types: list[QPUJobType]

    model_config = ConfigDict(from_attributes=True)


class QPUState(QPU):
    pending_internal_jobs: int | None = None
    pending_external_jobs: int | None = None


class Heartbeat(BaseModel):
    time_complete: str
    atom_number: int
    condensate_fraction: int
    temperature: float


class QPUStatusUpdate(BaseModel):
    status: QPUStatus

    # @root_validator()
    # def check_status_or_hours(cls, values):
    #     if (values.get("status") is None) and (values.get("operation_hours") is None):
    #         raise ValueError("either status or operation_hours is required")
    #     return values



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date, time
from enum import Enum

from pydantic import BaseModel

from bert_schemas import qpu


class DayOfWeek(str, Enum):
    MONDAY = "MONDAY"
    TUESDAY = "TUESDAY"
    WEDNESDAY = "WEDNESDAY"
    THURSDAY = "THURSDAY"
    FRIDAY = "FRIDAY"
    SATURDAY = "SATURDAY"
    SUNDAY = "SUNDAY"

    def __str__(self):
        return str(self.value)


class AccessType(str, Enum):
    GROUP = "GROUP"
    ROLE = "ROLE"
    ORG = "ORG"
    QPU = "QPU"

    def __str__(self):
        return str(self.value)


class AccessSlot(BaseModel):
    day: DayOfWeek
    start_date: date
    end_date: date | None = None
    start_time: time
    end_time: time


class Access(BaseModel):
    qpu_name: qpu.QPUName
    access_name: str | None = None
    access_type: AccessType
    access_slots: list[AccessSlot] = []



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for pydantic factories"""

# from ..schemas.job import BertJobSchema


def get_input_count(_, values: dict) -> int:
    return len(values["inputs"])


def add_notes(job):
    for number, input in enumerate(job.inputs):
        input.notes = f"test {job.job_type.lower()} note #{number + 1}"
    return job


def set_run(job):
    for i, input in enumerate(job.inputs):
        input.run = i + 1
    return job



from dataclasses import dataclass
from decimal import Decimal
from typing import Annotated

import numpy as np
from pydantic import Field, PlainSerializer


@dataclass(frozen=True)
class Settings:
    max_evap_duration: int = 2000  # maximum allowed rf evaporation duration
    max_exp_time: Decimal = Decimal("150.0")  # maximum allowed user experiment time
    min_tof: Decimal = Decimal("2.0")  # minimum allowed value for time-of-flight
    max_tof: Decimal = Decimal("20.0")  # maximum allowed value for time-of-flight


TimeMs = Annotated[
    Decimal,
    Field(ge=Decimal("0.0"), le=Settings.max_exp_time, decimal_places=1),
    PlainSerializer(lambda x: decimal_to_float(x), return_type=float),
]


EndTimeMs = TimeMs


TimeOfFlightMs = Annotated[TimeMs, Field(ge=Settings.min_tof, le=Settings.max_tof)]


def ceiling(x: float, precision=0):
    """Performs ceiling function with caller-specified precision

    Args:
        x (float): _description_
        precision (int, optional): Number of decimal points. Defaults to 0.

    Returns:
        float: output
    """
    return np.true_divide(np.ceil(x * 10**precision), 10**precision)


def floor(x: float, precision=0):
    """Performs ceiling function with caller-specified precision

    Args:
        x (float): _description_
        precision (int, optional): Number of decimal points. Defaults to 0.

    Returns:
        float: output
    """
    return np.true_divide(np.floor(x * 10**precision), 10**precision)


def resolve_time(t: float) -> float:
    """Resolves input time to nearest 0.1 ms, corresponding to the allowed time resolution.

    Args:
        t (float): input time in ms

    Returns:
        float: output time in ms
    """
    return floor(t, 1)


def resolve_times(ts: list) -> list:
    """Resolves input times to nearest 0.1 ms, corresponding to the allowed time resolution.

    Args:
        ts (list): input time in ms

    Returns:
        list: output times in ms
    """
    return [resolve_time(t) for t in ts]


def decimal_times_are_unique(ts: list[Decimal]) -> bool:
    """Tests if all object times are in fact, unique

    Args:
        ts (list): list of times

    Returns:
        bool: true if times are unique
    """
    float_times = decimals_to_floats(ts)
    return all(np.diff(float_times))


def decimal_to_float(d: Decimal) -> float:
    """Converts decimal time to float

    Args:
        d (Decimal): input time in ms
    Returns:
        float: output time in ms, resolved to 0.1 resolution
    """
    return resolve_time(d)


def decimals_to_floats(ds: list[Decimal]) -> list[float]:
    """Converts lists of decimal times to floats

    Args:
        ds (list): input times in ms

    Returns:
        list: output times in ms, resolved to 0.1 resolution
    """
    return resolve_times(ds)


def float_to_decimal(f: float) -> Decimal:
    """Converts float to decimal value with 0.1 resolution

    Args:
        ds (list): input times in ms

    Returns:
        list: output times in ms, resolved to 0.1 resolution
    """
    return Decimal(resolve_time(f))


def floats_to_decimals(fs: list[float]) -> list[Decimal]:
    """Converts lists of decimals to floats

    Args:
        fs (list): input times in ms, as floats

    Returns:
        list: output times in ms, resolved to 0.1 resolution
    """
    return [float_to_decimal(f) for f in fs]



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from pydantic import BaseModel, ConfigDict, field_validator


class MessageBase(BaseModel):
    location: int = 1
    message: str
    start_datetime: datetime
    end_datetime: datetime
    model_config = ConfigDict(from_attributes=True, validate_assignment=True)


class Message(MessageBase):
    id: int


class MessageResponse(Message):
    location: int = 1
    message: str
    start_datetime: str | datetime
    end_datetime: str | datetime

    @field_validator("start_datetime", mode="before")
    @classmethod
    def formate_start_datetime(cls, value):
        return value.strftime("%m/%d/%Y @ %H:%M")

    @field_validator("end_datetime", mode="before")
    @classmethod
    def formate_end_datetime(cls, value):
        return value.strftime("%m/%d/%Y @ %H:%M")
